@@ -26,12 +26,16 @@ use std::{
 
 use anyhow::{anyhow, Context, Error};
 use bollard::{
+    auth::DockerCredentials,
     container::{
         Config as BollardContainerConfig, CreateContainerOptions, ListContainersOptions,
         LogsOptions, StartContainerOptions, WaitContainerOptions,
     },
     exec::{CreateExecOptions, StartExecOptions},
-    models::{EndpointSettings, HostConfig, HostConfigLogConfig, PortBinding},
+    models::{
+        EndpointSettings, HostConfig, HostConfigLogConfig, PortBinding, RestartPolicy,
+        RestartPolicyNameEnum,
+    },
     network::{ConnectNetworkOptions, CreateNetworkOptions, ListNetworksOptions},
     Docker,
 };
@@ -71,6 +75,14 @@ lazy_static! {
 /// The amount of memory to allocate
 const MEMORY_ALLOCATION_BYTES: i64 = 4 * 1024 * 1024 * 1024;
 
+/// The maximal number of times we can restart Synapse in case it stops accidentally.
+///
+/// Accidental stops are typically due:
+/// 1. to Synapse not being able to open its port at startup (this happens, for reasons unknown);
+/// 2. to Synapse receiving a SIGTERM (this happens, for reasons unknown);
+/// 3. to a synax error or startup error in a module.
+const MAX_SYNAPSE_RESTART_COUNT: i64 = 20;
+
 /// A port in the container made accessible on the host machine.
 #[derive(Clone, Debug, Deserialize)]
 pub struct PortMapping {
@@ -123,11 +135,13 @@ impl DockerConfig {
 #[derive(Debug, Deserialize, Serialize, TypedBuilder)]
 pub struct HomeserverConfig {
     /// The name of the homeserver.
-    #[builder(default = "localhost:9999".to_string())]
+    #[serde(default = "HomeserverConfig::server_name_default")]
+    #[builder(default = HomeserverConfig::server_name_default())]
     pub server_name: String,
 
     /// The URL to communicate to the server with.
-    #[builder(default = "http://localhost:9999".to_string())]
+    #[serde(default = "HomeserverConfig::public_baseurl_default")]
+    #[builder(default = HomeserverConfig::public_baseurl_default())]
     pub public_baseurl: String,
 
     #[serde(default = "HomeserverConfig::registration_shared_secret_default")]
@@ -148,6 +162,12 @@ impl Default for HomeserverConfig {
 }
 
 impl HomeserverConfig {
+    pub fn server_name_default() -> String {
+        "localhost:9999".to_string()
+    }
+    pub fn public_baseurl_default() -> String {
+        format!("http://{}", Self::server_name_default())
+    }
     pub fn registration_shared_secret_default() -> String {
         "MX_TESTER_REGISTRATION_DEFAULT".to_string()
     }
@@ -198,6 +218,20 @@ pub struct Config {
     #[builder(default)]
     /// The version of Synapse to use
     pub synapse: SynapseVersion,
+
+    #[serde(default)]
+    #[builder(default)]
+    /// Information for logging to a registry.
+    ///
+    /// May be overridden from the command-line.
+    pub credentials: DockerCredentials,
+
+    #[serde(default)]
+    #[builder(default)]
+    /// Directories to use for the test.
+    ///
+    /// May be overridden from the command-line.
+    pub directories: Directories,
 }
 
 impl Config {
@@ -208,9 +242,9 @@ impl Config {
     /// script step.
     pub fn shared_env_variables(&self) -> Result<HashMap<&'static OsStr, OsString>, Error> {
         let synapse_root = self.synapse_root();
-        let script_tmpdir = std::env::temp_dir().join("mx-tester").join("scripts");
+        let script_tmpdir = self.directories.root.join("scripts");
         std::fs::create_dir_all(&script_tmpdir)
-            .with_context(|| format!("Could not create directory {:?}", script_tmpdir,))?;
+            .with_context(|| format!("Could not create directory {:#?}", script_tmpdir,))?;
         let curdir = std::env::current_dir()?;
         let mut env: HashMap<&'static OsStr, _> = HashMap::new();
         env.insert(&*MX_TEST_SYNAPSE_DIR, synapse_root.as_os_str().into());
@@ -224,6 +258,7 @@ impl Config {
     pub fn patch_homeserver_config(&self) -> Result<(), Error> {
         use serde_yaml::{Mapping, Value as YAML};
         let target_path = self.synapse_root().join("data").join("homeserver.yaml");
+        debug!("Attempting to open {:#?}", target_path);
         let config_file = std::fs::File::open(&target_path)
             .context("Could not open the homeserver.yaml generated by synapse")?;
         let mut combined_config: Mapping = serde_yaml::from_reader(config_file)
@@ -246,7 +281,7 @@ impl Config {
                 let listeners = listeners
                     .as_sequence_mut()
                     .ok_or_else(|| anyhow!("`listeners` should be a sequence"))?;
-                debug!("Listeners: {:?}", listeners);
+                debug!("Listeners: {:#?}", listeners);
 
                 // FIXME: For the time being, let's only handle the simplest case.
                 // If we're lucky, we'll be able to find WTH is going on that causes
@@ -307,7 +342,7 @@ impl Config {
     ///
     /// Cleaned up upon test start.
     pub fn test_root(&self) -> PathBuf {
-        std::env::temp_dir().join("mx-tester").join(&self.name)
+        self.directories.root.join(&self.name)
     }
 
     /// The directory in which we're putting synapse data for this test.
@@ -341,6 +376,23 @@ impl Config {
     }
 }
 
+/// Configurable directories for this test.
+#[derive(Debug, TypedBuilder, Deserialize)]
+pub struct Directories {
+    /// The root of the test.
+    ///
+    /// All temporary files and logs are created as subdirectories of this directory.
+    ///
+    /// If unspecified, `mx-tester` in the platform's temporary directory.
+    #[builder(default=std::env::temp_dir().join("mx-tester"))]
+    pub root: PathBuf,
+}
+impl Default for Directories {
+    fn default() -> Self {
+        Directories::builder().build()
+    }
+}
+
 /// The result of the test, as seen by `down()`.
 pub enum Status {
     /// The test was a success.
@@ -529,7 +581,15 @@ async fn start_synapse_container(
             .iter()
             .map(|mapping| (format!("{}/tcp", mapping.guest), HashMap::new()))
             .collect();
-        debug!("port_bindings: {:?}", host_port_bindings);
+        debug!("port_bindings: {:#?}", host_port_bindings);
+
+        debug!(
+            "containers: {:#?}",
+            docker
+                .list_containers(None::<ListContainersOptions<String>>)
+                .await?
+        );
+        assert!(std::fs::metadata(data_dir).is_ok());
 
         debug!("Creating container {}", container_name);
         let response = docker
@@ -546,16 +606,28 @@ async fn start_synapse_container(
                             typ: Some("json-file".to_string()),
                             config: None,
                         }),
+                        // Synapse has a tendency to not start correctly
+                        // or to stop shortly after startup. The following
+                        // restart policy seems to help a lot.
+                        restart_policy: Some(RestartPolicy {
+                            name: Some(RestartPolicyNameEnum::ON_FAILURE),
+                            maximum_retry_count: Some(MAX_SYNAPSE_RESTART_COUNT),
+                        }),
                         // Extremely large memory allowance.
                         memory_reservation: Some(MEMORY_ALLOCATION_BYTES),
                         memory_swap: Some(-1),
                         // Mount guest directory `/data` into the host synapse data directory.
                         binds: Some(vec![format!(
-                            "{}:/data",
+                            "{}:/data:rw",
                             data_dir.as_os_str().to_string_lossy()
                         )]),
                         // Expose guest port `guest_mapping` as `host_mapping`.
                         port_bindings: Some(host_port_bindings),
+                        // Enable access to host as `host.docker.internal` from the guest.
+                        // On macOS and Windows, this is expected to be transparent but
+                        // on Linux, an option needs to be added.
+                        #[cfg(target_os = "linux")]
+                        extra_hosts: Some(vec!["host.docker.internal:host-gateway".to_string()]),
                         ..HostConfig::default()
                     }),
                     image: Some(config.tag()),
@@ -564,6 +636,7 @@ async fn start_synapse_container(
                     attach_stdin: Some(false),
                     cmd: Some(cmd.clone()),
                     // Specify that `/data` may be mounted.
+                    // The empty hashmap... is an oddity of the Docker Engine API.
                     volumes: Some(
                         vec![("/data".to_string(), HashMap::new())]
                             .into_iter()
@@ -591,7 +664,7 @@ async fn start_synapse_container(
                 debug!(target: "mx-tester-wait", "{} Container started", container_name);
                 while let Some(next) = wait.next().await {
                     let response = next.context("Error while waiting for container to stop")?;
-                    debug!(target: "mx-tester-wait", "{} {:?}", container_name, response);
+                    debug!(target: "mx-tester-wait", "{} {:#?}", container_name, response);
                 }
                 debug!(target: "mx-tester-wait", "{} Container is now down", container_name);
                 Ok::<(), Error>(())
@@ -633,6 +706,7 @@ async fn start_synapse_container(
         );
 
         // Write logs to the synapse data directory.
+
         let mut log_file = tokio::fs::OpenOptions::new()
             .create(true)
             .append(true)
@@ -729,7 +803,7 @@ pub async fn build(docker: &Docker, config: &Config) -> Result<(), Error> {
     let synapse_root = config.synapse_root();
     let _ = std::fs::remove_dir_all(config.test_root());
     std::fs::create_dir_all(&synapse_root)
-        .with_context(|| format!("Could not create directory {:?}", synapse_root,))?;
+        .with_context(|| format!("Could not create directory {:#?}", synapse_root,))?;
 
     // Build modules
     let mut env = config.shared_env_variables()?;
@@ -737,7 +811,7 @@ pub async fn build(docker: &Docker, config: &Config) -> Result<(), Error> {
         let path = synapse_root.join(&module.name);
         env.insert(&*MX_TEST_MODULE_DIR, path.as_os_str().into());
         debug!(
-            "Calling build script for module {} with MX_TEST_DIR={:?}",
+            "Calling build script for module {} with MX_TEST_DIR={:#?}",
             &module.name, path
         );
         module
@@ -753,16 +827,13 @@ pub async fn build(docker: &Docker, config: &Config) -> Result<(), Error> {
 
 FROM {docker_tag}
 
-# We need gcc to build pyahocorasick
-RUN apt-get update --quiet && apt-get install gcc --yes --quiet
-
 # Show the Synapse version, to aid with debugging.
 RUN pip show matrix-synapse
 
 # Copy and install custom modules.
 RUN mkdir /mx-tester
-{copy}
 {setup}
+{copy}
 
 VOLUME [\"/data\"]
 ENTRYPOINT []
@@ -770,31 +841,31 @@ ENV SYNAPSE_HTTP_PORT=8008
 EXPOSE 8008/tcp 8009/tcp 8448/tcp
 ",
     docker_tag = docker_tag,
+    setup = config.modules.iter()
+        .filter_map(|module| module.install.as_ref().map(|script| format!("## Setup {}\n{}\n", module.name, script.lines.iter().map(|line| format!("RUN {}", line)).format("\n"))))
+        .format("\n"),
     copy = config.modules.iter()
         // FIXME: We probably want to test what happens with weird characters. Perhaps we'll need to somehow escape module.
         .map(|module| format!("COPY {module} /mx-tester/{module}\nRUN /usr/local/bin/python -m pip install /mx-tester/{module}", module=module.name))
         .format("\n"),
-    setup = config.modules.iter()
-        .filter_map(|module| module.install.as_ref().map(|script| format!("## Install {}\n{}\n", module.name, script.lines.iter().map(|line| format!("RUN {}", line)).format("\n"))))
-        .format("\n")
-);
+    );
     debug!("dockerfile {}", dockerfile_content);
 
     let dockerfile_path = synapse_root.join("Dockerfile");
     std::fs::write(&dockerfile_path, dockerfile_content)
-        .with_context(|| format!("Could not write file {:?}", dockerfile_path,))?;
+        .with_context(|| format!("Could not write file {:#?}", dockerfile_path,))?;
 
     debug!("Building tar file");
-    let docker_dir_path = std::env::temp_dir().join("mx-tester").join("tar");
+    let docker_dir_path = config.test_root().join("tar");
     std::fs::create_dir_all(&docker_dir_path)
-        .with_context(|| format!("Could not create directory {:?}", docker_dir_path,))?;
+        .with_context(|| format!("Could not create directory {:#?}", docker_dir_path,))?;
     let body = {
         // Build the tar file.
         let tar_path = docker_dir_path.join("docker.tar");
         {
             let tar_file = std::fs::File::create(&tar_path)?;
             let mut tar_builder = tar::Builder::new(tar_file);
-            debug!("tar: adding directory {:?}", synapse_root);
+            debug!("tar: adding directory {:#?}", synapse_root);
             tar_builder.append_dir_all("", &synapse_root)?;
             tar_builder.finish()?;
         }
@@ -814,19 +885,23 @@ EXPOSE 8008/tcp 8009/tcp 8448/tcp
                 rm: true,
                 ..Default::default()
             },
-            None,
+            config.credentials.serveraddress.as_ref().map(|server| {
+                let mut credentials = HashMap::new();
+                credentials.insert(server.clone(), config.credentials.clone());
+                credentials
+            }),
             Some(body),
         );
         while let Some(result) = stream.next().await {
             let info = result.context("Daemon `docker build` indicated an error")?;
             if let Some(ref error) = info.error {
                 return Err(anyhow!(
-                    "Error while building an image {}: {:?}",
+                    "Error while building an image {}: {:#?}",
                     error,
                     info.error_detail
                 ));
             }
-            debug!("Build image progress {:?}", info);
+            debug!("Build image progress {:#?}", info);
         }
     }
     debug!("Image built");
@@ -837,9 +912,30 @@ EXPOSE 8008/tcp 8009/tcp 8448/tcp
 pub async fn up(docker: &Docker, config: &Config) -> Result<(), Error> {
     // This will break (on purpose) once we extend `SynapseVersion`.
     let SynapseVersion::Docker { .. } = config.synapse;
-    let setup_container_name = config.setup_container_name();
-    let run_container_name = config.run_container_name();
 
+    // Create the network if necessary.
+    // We'll add the container once it's available.
+    let network_name = config.network();
+    debug!("We'll need network {}", network_name);
+    if !docker.is_network_up(&network_name).await? {
+        debug!("Creating network {}", network_name);
+        docker
+            .create_network(CreateNetworkOptions {
+                name: network_name.as_str(),
+                ..CreateNetworkOptions::default()
+            })
+            .await?;
+        assert!(
+            docker.is_network_up(&network_name).await?,
+            "The network should now be up"
+        );
+    } else {
+        debug!("Network {} already exists", network_name);
+    }
+
+    // Only execute the `up` script once the network is up,
+    // in case we want to e.g. bring up images that need
+    // that same network.
     match config.up {
         Some(UpScript::FullUpScript(FullUpScript {
             before: Some(ref script),
@@ -854,27 +950,15 @@ pub async fn up(docker: &Docker, config: &Config) -> Result<(), Error> {
         _ => {}
     }
 
-    // Create the network if necessary.
-    // We'll add the container once it's available.
-    let network_name = config.network();
-    if !docker.is_network_up(&network_name).await? {
-        docker
-            .create_network(CreateNetworkOptions {
-                name: network_name.as_str(),
-                ..CreateNetworkOptions::default()
-            })
-            .await?;
-        assert!(
-            docker.is_network_up(&network_name).await?,
-            "The network should now be up"
-        );
-    }
+    let setup_container_name = config.setup_container_name();
+    let run_container_name = config.run_container_name();
 
     // Create the synapse data directory.
     // We'll use it as volume.
     let synapse_data_directory = config.synapse_root().join("data");
     std::fs::create_dir_all(&synapse_data_directory)
-        .with_context(|| format!("Cannot create directory {:?}", synapse_data_directory))?;
+        .with_context(|| format!("Cannot create directory {:#?}", synapse_data_directory))?;
+
     // Cleanup leftovers.
     let homeserver_path = synapse_data_directory.join("homeserver.yaml");
     let _ = std::fs::remove_file(&homeserver_path);
@@ -900,12 +984,30 @@ pub async fn up(docker: &Docker, config: &Config) -> Result<(), Error> {
     let _ = docker.stop_container(&setup_container_name, None).await;
     let _ = docker.remove_container(&setup_container_name, None).await;
 
+    debug!("Updating homeserver.yaml");
     // Apply config from mx-tester.yml to the homeserver.yaml that was just created
     config
         .patch_homeserver_config()
         .context("Error updating homeserver config")?;
 
-    // It's now time to run Synapse.
+    // Docker has a tendency to return before containers are fully torn down.
+    // Let's make extra-sure by waiting until the container is not running
+    // anymore *and* the ports are free.
+    while docker.is_container_running(&setup_container_name).await? {
+        debug!(
+            "Waiting until docker container {} is down before relaunching it",
+            setup_container_name
+        );
+        tokio::time::sleep(std::time::Duration::new(5, 0)).await;
+    }
+    while let Some(mapping) = config.docker.port_mapping.iter().find(|mapping| {
+        debug!("Checking whether port {} is available", mapping.host);
+        std::net::TcpListener::bind(("127.0.0.1", mapping.host as u16)).is_err()
+    }) {
+        debug!("Waiting until port {} is available", mapping.host);
+        tokio::time::sleep(std::time::Duration::new(5, 0)).await;
+    }
+    debug!("Port is available, proceeding");
     start_synapse_container(
         docker,
         config,
@@ -917,10 +1019,36 @@ pub async fn up(docker: &Docker, config: &Config) -> Result<(), Error> {
     .await
     .context("Failed to start Synapse")?;
 
-    debug!("Synapse is now launched");
-    handle_user_registration(config)
-        .await
-        .context("Failed to setup users")?;
+    debug!("Synapse should now be launched and ready");
+
+    // We should now be able to register users.
+    //
+    // As of this writing, we're not sure whether the `synapse_is_responsive` manipulation
+    // above works. If it doesn't, we can still have a case in which Synapse won't start,
+    // causing `handle_user_registration` to loop endlessly. The `timeout` should make
+    // sure that we fail properly and with an understandable error message.
+    //
+    // This will presumably disappear if the `synapse_is_responsive` manipulation above works.
+    match tokio::time::timeout(std::time::Duration::new(120, 0), async {
+        handle_user_registration(config)
+            .await
+            .context("Failed to setup users")
+    })
+    .await
+    {
+        Err(_) => {
+            // Timeout.
+            panic!(
+                "User registration is taking too long. {}",
+                if docker.is_container_running(&run_container_name).await? {
+                    "Container is running."
+                } else {
+                    "For some reason, Synapse has stopped. Please check the Synapse logs and/or rerun `mx-tester up`."
+                }
+            );
+        }
+        Ok(result) => result,
+    }?;
     if let Some(UpScript::FullUpScript(FullUpScript {
         after: Some(ref script),
         ..
@@ -940,7 +1068,9 @@ pub async fn down(docker: &Docker, config: &Config, status: Status) -> Result<()
     let SynapseVersion::Docker { .. } = config.synapse;
     let run_container_name = config.run_container_name();
 
-    if let Some(ref down_script) = config.down {
+    // Store results, we'll report them after we've brought down everything
+    // that we can bring down.
+    let script_result = if let Some(ref down_script) = config.down {
         let env = config.shared_env_variables()?;
         // First run on_failure/on_success.
         // Store errors for later.
@@ -967,31 +1097,59 @@ pub async fn down(docker: &Docker, config: &Config, status: Status) -> Result<()
         };
         // Then run on_always.
         if let Some(ref on_always) = down_script.finally {
-            on_always
-                .run(&env)
-                .context("Error while running script `down/finally`")?;
+            result.and(
+                on_always
+                    .run(&env)
+                    .context("Error while running script `down/finally`"),
+            )
+        } else {
+            result
         }
-        // Report any error from `on_failure` or `on_success`.
-        result?
-    }
+    } else {
+        Ok(())
+    };
+
     debug!(target: "mx-tester-down", "Taking down synapse.");
-    match docker.stop_container(&run_container_name, None).await {
+    let stop_container_result = match docker.stop_container(&run_container_name, None).await {
         Err(bollard::errors::Error::DockerResponseNotModifiedError { .. }) => {
             // Synapse is already down.
             debug!(target: "mx-tester-down", "Synapse was already down");
+            Ok(())
         }
         Err(bollard::errors::Error::DockerResponseNotFoundError { .. }) => {
             // Synapse is already down.
             debug!(target: "mx-tester-down", "No Synapse container");
+            Ok(())
         }
         Ok(_) => {
             debug!(target: "mx-tester-down", "Synapse taken down");
+            Ok(())
         }
-        Err(err) => {
-            return Err(err).context("Error stopping container");
+        Err(err) => Err(err).context("Error stopping container"),
+    };
+
+    debug!(target: "mx-tester-down", "Taking down network.");
+    let remove_network_result = match docker.remove_network(config.network().as_ref()).await {
+        Err(bollard::errors::Error::DockerResponseNotModifiedError { .. }) => {
+            // Network is already down.
+            debug!(target: "mx-tester-down", "Network was already down");
+            Ok(())
         }
-    }
-    Ok(())
+        Err(bollard::errors::Error::DockerResponseNotFoundError { .. }) => {
+            // Network is already down.
+            debug!(target: "mx-tester-down", "No network");
+            Ok(())
+        }
+        Ok(_) => {
+            debug!(target: "mx-tester-down", "Network taken down");
+            Ok(())
+        }
+        Err(err) => Err(err).context("Error stopping network"),
+    };
+    // Finally, report any problem.
+    script_result
+        .and(stop_container_result)
+        .and(remove_network_result)
 }
 
 /// Run the testing script.
@@ -1025,7 +1183,7 @@ impl DockerExt for Docker {
                 filters: vec![("name", vec![name])].into_iter().collect(),
             }))
             .await?;
-        debug!("is_network_up {:?}", networks);
+        debug!("is_network_up {:#?}", networks);
         Ok(networks.is_empty().not())
     }
 
@@ -1040,7 +1198,7 @@ impl DockerExt for Docker {
                 ..ListContainersOptions::default()
             }))
             .await?;
-        debug!("is_container_running {:?}", containers);
+        debug!("is_container_running {:#?}", containers);
         Ok(containers.is_empty().not())
     }
 
@@ -1055,7 +1213,7 @@ impl DockerExt for Docker {
                 ..ListContainersOptions::default()
             }))
             .await?;
-        debug!("is_container_created {:?}", containers);
+        debug!("is_container_created {:#?}", containers);
         Ok(containers.is_empty().not())
     }
 }
